#include <iostream>
#include <fstream>
#include <string.h>
#include <string>
#include <vector>
#include <map>
#include <algorithm>

//#define NDEBUG
#include <cassert>

#define MAXINDEX 1000

using namespace std;

struct Align {int s_i; int t_i;};

int LoadSnt (const string& path, vector <vector <string> >& snt)
{
	cerr << "load sentences from " << path << " ... \n";
	ifstream snt_file (path.c_str());
	if (!snt_file)
	{
		cerr << "cannot find " << path << "!!! \n";
		exit (1);
	}
	string line;
	while (getline (snt_file, line))
	{
		int prev = 0;
		int post = 0;
		vector <string> seg;
		while (true)
		{
			for (; (line [prev] != ' ' && line [prev] != '\t') && (prev < line.length ()); prev ++) {}
			seg.push_back (line.substr (post, prev-post));
			prev ++;
			post = prev;
			if (prev >= line.length ()) break;
		}
		snt.push_back (seg);
	}
	snt_file.close ();
	return 0;
}

int LoadRI (const string& path, vector <vector <int> >& ri)
{
	cerr << "load reversed-index lines from " << path << " ... \n";
	ifstream ri_file (path.c_str());
	if (!ri_file)
	{
		cerr << "cannot find " << path << "!!! \n";
		exit (1);
	}
	string line;
	while (getline (ri_file, line))
	{
		int prev = 0;
		int post = 0;
		vector <int> seg;
		while (true)
		{
			for (; (line [prev] != ' ' && line [prev] != '\t') && (prev < line.length ()); prev ++) {}
			seg.push_back (atoi (line.substr (post, prev-post).c_str ()));
			prev ++;
			post = prev;
			if (prev >= line.length ()) break;
		}
		ri.push_back (seg);
	}
	ri_file.close ();
	return 0;
}

int LoadAlign (const string& path, vector <vector <Align> >& align)
{
	cerr << "load alignment lines from " << path << " ... \n";
	ifstream a_file (path.c_str());
	if (!a_file)
	{
		cerr << "cannot find " << path << "!!! \n";
		exit (1);
	}
	string line;
	while (getline (a_file, line))
	{
		int prev = 0;
		int post = 0;
		vector <string> seg;
		while (true)
		{
			for (; (line [prev] != ' ' && line [prev] != '\t') && (prev < line.length ()); prev ++) {}
			seg.push_back (line.substr (post, prev-post));
			prev ++;
			post = prev;
			if (prev >= line.length ()) break;
		}
		vector <Align> aseg;
		for (int i = 0; i < seg.size (); i ++)
		{
	                int pos = 0;
	                for (; (seg [i][pos] != '-') && (pos < seg [i].length ()); pos ++) {}
	                int s = atoi (seg [i].substr (0, pos).c_str());
	                int t = atoi (seg [i].substr (pos+1, line.length ()-pos-1).c_str());
	                Align new_align;
	                new_align.s_i = s;
	                new_align.t_i = t;
	                aseg.push_back (new_align);			
		}
		align.push_back (aseg);
        }
	a_file.close ();
	return 0;
}

int RecRev (const vector <string>& s_snt, const vector <int>& i_line, const vector <Align>& a_line, \
		vector <int>& r_i, vector <string>& r_s, vector <Align>& r_a)
{
	int smax = s_snt.size ();
	vector <vector <int> > s_t;
	for (int i = 0; i < smax; i ++) {vector <int> new_vec; s_t.push_back (new_vec);}
	for (int i = 0; i < a_line.size (); i ++)
	{
		int s = a_line [i].s_i;
		int t = a_line [i].t_i;
		assert (s >= 0 && s <= smax);
		s_t [s].push_back (t);
	}
	vector <int> rev_ori;
	for (int i = 0; i < smax; i ++) rev_ori.push_back (-1);
	for (int i = 0; i < smax; i ++) rev_ori [i_line [i]] = i;
	for (int i = 0; i < smax; i ++) assert (!(rev_ori [i] < 0));
	for (int i = 0; i < smax; i ++) r_i.push_back (rev_ori [i]);
	for (int i = 0; i < smax; i ++) r_s.push_back (s_snt [rev_ori [i]]);
	for (int i = 0; i < smax; i ++)
		for (int j = 0; j < s_t [i].size (); j ++)
		{
			Align new_align;
			new_align.s_i = rev_ori [i];
			new_align.t_i = s_t [i][j];
			r_a.push_back (new_align);
		}
	assert (r_a.size () == a_line.size ());
	
	return 0;
}

int SegRev (const vector <string>& s_snt, const vector <string>& t_snt, const vector <Align>& a_line, \
		vector <int>& r_i, vector <string>& r_s, vector <Align>& r_a)
{
	int smax = s_snt.size ();
	int tmax = t_snt.size ();
	string label = "";
	for (int i = 0; i < smax; i ++) label += '-';
	vector <vector <int> > s_t;
	vector <vector <int> > t_s;
	for (int i = 0; i < smax; i ++) {vector <int> new_vec; s_t.push_back (new_vec);}
	for (int i = 0; i < tmax; i ++)	{vector <int> new_vec; t_s.push_back (new_vec);}
	for (int i = 0; i < a_line.size (); i ++)
	{
		int s = a_line [i].s_i;
		int t = a_line [i].t_i;
		assert (s >= 0 && s <= smax);
		assert (t >= 0 && t <= tmax);
		s_t [s].push_back (t);
		t_s [t].push_back (s);
	}
	int sstart = 0;
	int pre_tend = -1;
	for (int send = sstart; send < smax; send ++)
	{
		bool good = true;
		int tstart = tmax;
		int tend = -1;
		for (int s = sstart; s <= send ; s ++)
			for (int t = 0; t < s_t [s].size (); t ++)
			{
				if (s_t [s][t] < tstart) tstart = s_t [s][t];
				if (s_t [s][t] > tend) tend = s_t [s][t];
			}
		for (int t = pre_tend + 1; t < tstart; t ++)
			if (t_s [t].size () != 0) {good = false; break;}
		if (!good) continue;
		for (int t = tstart; t <= tend; t ++)
			for (int s = 0; s < t_s [t].size (); s ++)
			if (t_s [t][s] < sstart || t_s [t][s] > send) {good = false; break;}
		if (!good) continue;
		pre_tend = tend;
		sstart = send + 1;
		label [send] = '|';
		while (sstart < smax && s_t [sstart].size () == 0) {label [sstart ++] = '*';}
	}
	//cerr << label << "\n";
	int start = 0;
	for (int i = 0; i < smax; i ++)
	{
		if (label [i] != '|') continue;
		int con = 0;
		int dis = 0;
		for (int j = start; j <= i; j ++)
			for (int k = j; k <= i; k ++)
				for (int t1 = 0; t1 < s_t [j].size (); t1 ++)
					for (int t2 = 0; t2 < s_t [k].size (); t2 ++)
					{
						if	(s_t [j][t1] < s_t [k][t2]) con ++;
						else if	(s_t [j][t1] > s_t [k][t2]) dis ++;
					}
		if (dis > con) label [i] = '<';	
		start = i + 1;
	}
	//cerr << label << "\n";
	start = 0;
	for (int i = 0; i < smax; i ++)
	{
		if (label [i] == '-') continue;
		if (label [i] == '*') {start = i + 1; r_i.push_back (i); continue;}
		if 	(label [i] == '|')
		{
			for (int j = start; j <= i; j ++) r_i.push_back (j);
			start = i + 1;
		}
		else if (label [i] == '<')
		{
			for (int j = i; j >= start; j --) r_i.push_back (j);
			start = i + 1;
		}
		else assert (false); 
	}
	//for (int i = 0; i < r_i.size (); i ++) cerr << r_i [i] << " ";
	//cerr << "\n";
	assert (r_i.size () == smax);
	vector <int> ori_rev;
	for (int i = 0; i < smax; i ++) ori_rev.push_back (-1);
	for (int i = 0; i < smax; i ++) ori_rev [r_i [i]] = i;
	for (int i = 0; i < smax; i ++) assert (!(ori_rev [i] < 0));
	for (int i = 0; i < smax; i ++) r_s.push_back (s_snt [ori_rev [i]]);
	for (int i = 0; i < smax; i ++)
		for (int j = 0; j < s_t [i].size (); j ++)
		{
			Align new_align;
			new_align.s_i = ori_rev [i];
			new_align.t_i = s_t [i][j];
			r_a.push_back (new_align);
		}
	assert (r_a.size () == a_line.size ());
	//for (int i = 0; i < smax; i ++) cerr << s_snt [i] << " ";
	//cerr << "\n";
	//for (int i = 0; i < smax; i ++) cerr << r_s [i] << " ";
	//cerr << "\n";
	//for (int i = 0; i < tmax; i ++) cerr << t_snt [i] << " ";
	//cerr << "\n\n";

	
	return 0;
}

int main (int argc, char* argv [])
{
	string S_Path = "";
	string T_Path = "";
	string I_Path = "";
	string A_Path = "";
	string RI_File = "";
	string RS_File = "";
	string RA_File = "";
	bool to_recover = false;
	vector <vector <string> > S_Snt;
	vector <vector <string> > T_Snt;
	vector <vector <int> > I_Line;
	vector <vector <Align> > A_Line;
	vector <string> opt;
	for (int i = 0; i < argc; i ++)
	{
		string str;
		opt.push_back (str.assign (argv [i], strlen (argv [i])));
	}

	for (int i = 0; i < argc; i ++)
	{
		if (opt [i] == "-s") S_Path = opt [i+1];
		if (opt [i] == "-t") T_Path = opt [i+1];
		if (opt [i] == "-i") I_Path = opt [i+1];
		if (opt [i] == "-a") A_Path = opt [i+1];

		if (opt [i] == "-ri") RI_File = opt [i+1];
		if (opt [i] == "-rs") RS_File = opt [i+1];
		if (opt [i] == "-ra") RA_File = opt [i+1];

		if (opt [i] == "-rec") to_recover = true;
	}

	LoadSnt (S_Path, S_Snt);
	cerr << S_Snt.size () << " source side sentences loaded, ";
	int wordcounter = 0;
	for (int i = 0; i < S_Snt.size (); i ++) wordcounter += S_Snt [i].size ();
	cerr << wordcounter << " words ... \n";
	
	if (to_recover)
	{
		LoadRI (I_Path, I_Line);
		cerr << I_Line.size () << " reversed-index lines loaded, ";
		wordcounter = 0;
		for (int i = 0; i < I_Line.size (); i ++) wordcounter += I_Line [i].size ();
		cerr << wordcounter << " words ... \n";
	}
	else
	{
		LoadSnt (T_Path, T_Snt);
		cerr << T_Snt.size () << " target side sentences loaded, ";
		wordcounter = 0;
		for (int i = 0; i < T_Snt.size (); i ++) wordcounter += T_Snt [i].size ();
		cerr << wordcounter << " words ... \n";
	}

	LoadAlign (A_Path, A_Line);
	cerr << A_Line.size () << " alignment lines loaded, ";
	wordcounter = 0;
	for (int i = 0; i < A_Line.size (); i ++) wordcounter += A_Line [i].size ();
	cerr << wordcounter << " aligned word pairs ... \n";

	assert (S_Snt.size () == A_Line.size ());
	if (T_Snt.size () != 0) assert (T_Snt.size () == A_Line.size ());
	if (I_Line.size () != 0) assert (I_Line.size () == A_Line.size ());

	ofstream ri_file (RI_File.c_str());
	if (!ri_file)
	{
		cerr << "cannot write " << RI_File << "!!! \n";
		exit (1);
	}
	ofstream rs_file (RS_File.c_str());
	if (!rs_file)
	{
		cerr << "cannot write " << RS_File << "!!! \n";
		exit (1);
	}
	ofstream ra_file (RA_File.c_str());
	if (!ra_file)
	{
		cerr << "cannot write " << RA_File << "!!! \n";
		exit (1);
	}

	int MAXLINE = A_Line.size ();
	for (int i = 0; i < MAXLINE; i ++)
	{
		cerr << i+1 << " ... \r";
		vector <int>	r_i;
		vector <string>	r_s;
		vector <Align>	r_a;
		if (to_recover)	RecRev (S_Snt [i], I_Line [i], A_Line [i], r_i, r_s, r_a);
		else		SegRev (S_Snt [i], T_Snt [i], A_Line [i], r_i, r_s, r_a);

		for (int i = 0; i < r_i.size (); i ++) ri_file << r_i [i] << " ";
		ri_file << "\n";
		for (int i = 0; i < r_s.size (); i ++) rs_file << r_s [i] << " ";
		rs_file << "\n";
		for (int i = 0; i < r_a.size (); i ++) ra_file << r_a [i].s_i << "-" << r_a [i].t_i << " ";
		ra_file << "\n";
	}
	cerr << "\ndone.\n";

	ri_file.close ();
	rs_file.close ();
	ra_file.close ();

	return 0;
}
